#!/usr/bin/env python
#
import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gimpshelf import shelf
import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################
# -------------------------------------------------------------------------------------
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>FaceMaker'
#
# The first option creates new letter faces from over-sized visible layer/s
# The second option clears the contents of the old character layer before copying and pasting the new letter face
# The third and fourth options set the mode (legacy and new) to be used when copying and pasting to generate the new font face
# The fifth option removes non-font character layers from the image - apart from any layer name containing the word 'legend'
# The sixth option will re-position all layers to the origin of their parent image (Option 4 does this too)
# The seventh option will re-order the layers in the custom font file according to the following order
#		space.png (or  .ping)
#		Capital letters (A-Z)
#		Lower case letters (a-z)
#		Numbers (0-9)
#		Basic maths characters
# 		Punctuation characters
#		Other keyboard characters
#		Special characters (Pound Sterling sign, Copyright Sign, Registered Trademark sign)
# You can change this order by altering the appropriate lines (currently beginning at line no 130) of the plug-in
# The eighth option will display a message box listing all the characters (ie layers) defined in the custom font
# The ninth option defines the string to be used to append to single letter layer and/or path names
# The tenth option will add .png to all the layers with one character names ie  A to A.png
# The eleventh option will add .png to all the paths with one character names
# The twelfth option will combine the characters from the xcf custom font file selected, with the file selection tool below, with the current characters
# The thirteenth option is the xcf file selection tool
# The fourteenth option trims pixels from the character

# Version 0.01 Fixed adding .png to single letter layers and paths with special characters ie unicode characters
# Version 1.02 Adds an option to create a new face for a custom font from additional/decorative visible layers
# Version 1.03 Re-position space layer, if one exists, along with other letter layers and adds the option to remove non-font-character layers from the image
# Version 1.04 Trim and feather pixels from letter faces to remove halo effect
# Version 1.05 Fixed bug where character layer is not at image origin (and so doesn't copy new face correctly): Adds option to use different modes when copying from visible layers to font face
# Version 1.06 Added option to re-position all layers to the origin of their parent image ie with x and y co-ordinates set to zero: Re-organized dialog and filter structure so that all options can be set without any issues
# Version 1.07 Added option to clear the existing letter face before copying: Re-ordered flow of filter (yet again); Added options to combine two custom font files
# Version 1.08 Fixed potential bug when layer is bigger than image
#              Changed filter location
# Version 1.09 Added mode choice when combining two custom fonts
# Version 1.10 Added fixed position option for creating new characters from a visible layer
# Version 1.11 Adds an option to trim pixels from each character; changed plug-in name to Custom Font Layer Utilities
# Version 1.12 Adds an option to add lower-case characters to the custom font by copying and scaling, as a percentage, the upper-case characters (when the custom font has a path called base_line.png)
# Version 1.13 Adds an option to sandwich a character between a top layer and a bottom layer - optionally setting the layer size to the full width of the font image size
# Version 0.01 Custom Font Face Maker 
#              Removed options not used in making character faces from visible layer
#              Added option for using all non-character layers to make new character faces from sequentially
#              Added option to use non-character layers randomly
#              Added option to use non-character layers with groups according to non-character layer name (upper/lower/number/punctuation/other)
# Version 0.02 adds checking for unicode symbols, reworked using character group faces
# Version 0.03 general improvements
# Version 0.04 Re-worked use of random character faces
# Version 0.05 Added use of smaller increments for growing the letter size
# Version 0.06 Re-worked trimming halo from character outlines
# Version 0.07 Fixed mode issue
# Version 0.08 general improvements
# Version 0.09 Added opacity slider
# Version 0.10 Added preview to new image option, character group option and fixed a bug that had been introduced when copying characters from another file with fixed position set
# Version 0.11 Tabbed
#              Fixed fixed position bug in combine fonts mode
#              Fixed applying mode when merging characters
#              Added drop shadow to upper layer when combining two custom fonts
#              Added Tabbed dialog using Gimpui 3 by Arakne and made available by MareroQ


# Some global colours we may need to use

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkgrey = gimpcolor.RGB(50, 50, 50)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	


def do_stuff(image, layer, fixed_position, grow, many_faces, many_rnd, grouping, clear_contents, old_mode, new_mode, opacity, 
            do_merge_files, xcffilename, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, do_preview, ptab):

    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    pdb.gimp_context_set_feather(False)
    #pdb.gimp_context_set_antialias(False)
	
    do_sandwich = False
    add_case = False
    do_layer_names = False
    do_path_names = False
    do_sort = False
    do_place_at_origin = False
    do_remove_non_letter_layers = False
    do_layer_info = False
    trim = 0
	
    add_drop = False
    if (drop_colour>0 and do_merge_files): add_drop = True

    do_new_letters = True
    if do_merge_files: 
        do_new_letters = False
        fixed_position = False
	
    if do_new_letters:
	
	
        if do_preview:
		
            preview_other, preview_case, preview_shape, preview_vowel = 0, 0, 0, 0
		
            preview_im_width, preview_im_height = image.width, image.height
            preview_image = pdb.gimp_image_new(preview_im_width, preview_im_height, 0)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_im_width, preview_im_height, 1, "temp preview layer", 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            pdb.gimp_image_remove_layer(preview_image, temp_preview_layer)
            preview_display = pdb.gimp_display_new(preview_image)
            # Set start undo point for preview image
            pdb.gimp_image_undo_group_start(preview_image)	
            
		
        # define some character groupings
        upper_case, lower_case, number_case, punctuation_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz", "0123456789", ".,:;'!?\""
        vowel_case, consonant_case = "AEIOUaeiou","BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz"
        straight_case, round_case, mixed_case = "AEFHIKLMNTVWXYZikltvwxz", "CGOQSaceos", "BDJPRUbdfghjmnpqruy"
		
		
        # set groups
        if grouping == 0:
            group = "none"
        elif grouping == 1:
            group = "case"
        elif grouping == 2:
            group = "shape"
        elif grouping == 3:
            group = "vowel"
		
        # define some preview character values
        preview_layer_pos = -1
        preview_upper_pos = -1
        preview_lower_pos = -1
        preview_number_pos = -1
        preview_punctuation_pos = -1
        preview_vowel_pos = -1
        preview_consonant_pos = -1
        preview_straight_pos = -1
        preview_round_pos = -1
        preview_mixed_pos = -1
		
        mode = fn_get_mode(old_mode, new_mode)
        im_width = image.width
        im_height = image.height
    
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
				
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        multi_layer_list = "No non-character layers."
		
        num_multi_layers = 0
        while visi_loop < num_layers:# loop through layers and store any non-character layers for creating new faces
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # check for custom font layer now allows unicode symbols
                if first_letter_layer < 0: first_letter_layer = visi_loop
            else:
                pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
                is_empty = pdb.gimp_selection_is_empty(image)
                if is_empty == False:
                    if other_layers == 0:
                        
                        multi_layer_list = [visi_name]
                    else:
                        
                        multi_layer_list.append(visi_name)
                    other_layers += 1
                    
                    		
            visi_loop += 1

        if str(multi_layer_list) == "No non-character layers.": 
            num_multi_layers = 0
        else:
            num_multi_layers = len(multi_layer_list)
        if many_faces:
            if num_multi_layers < 2: many_faces = False
			
        if many_faces: copy_multi_layer_list = fn_copy_list(multi_layer_list)
			
        visi_loop = 0
        real_name_layer_list = ["dummy"] # list to hold actual character names 
        while visi_loop < num_layers and other_layers > 0: # turn off visibility of letter layers
            real_visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = real_visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": 
                has_a_space = True
                space_loop = visi_loop
            if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) 
                if real_visi_name<>"space.png" and real_visi_name<>" .png": # add some preview image variables
                    real_name_layer_list.append(real_visi_name)
                    if preview_layer_pos == -1: preview_layer_pos = visi_loop
                    if real_visi_name[0] in upper_case:
                        if preview_upper_pos == -1: preview_upper_pos = visi_loop
                    if real_visi_name[0] in lower_case:
                        if preview_lower_pos == -1: preview_lower_pos = visi_loop
                    if real_visi_name[0] in number_case:
                        if preview_number_pos == -1: preview_number_pos = visi_loop
                    if real_visi_name[0] in punctuation_case:
                        if preview_punctuation_pos == -1: preview_punctuation_pos = visi_loop
                    if real_visi_name[0] in vowel_case:
                        if preview_vowel_pos == -1: preview_vowel_pos = visi_loop
                    if real_visi_name[0] in consonant_case:
                        if preview_consonant_pos == -1: preview_consonant_pos = visi_loop
                    if real_visi_name[0] in straight_case:
                        if preview_straight_pos == -1: preview_straight_pos = visi_loop
                    if real_visi_name[0] in round_case:
                        if preview_round_pos == -1: preview_round_pos = visi_loop
                    if real_visi_name[0] in mixed_case:
                        if preview_mixed_pos == -1: preview_mixed_pos = visi_loop
					
            visi_loop += 1
        #real_name_layer_list.remove("dummy")
        #pdb.gimp_message(str(real_name_layer_list) + " preview_punctuation_pos = " + str(preview_punctuation_pos) + " layer name = " + str(real_name_layer_list[preview_punctuation_pos]) )
			
        if other_layers > 0: # start copying decor onto characters
            
            if do_preview:
                ### check for layers to copy to preview image/insert layers in preview layer/turn off layer visibility
                if group == "none":
                    preview_layer = layer_names_list[preview_layer_pos]
                    preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                    temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                    pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
                    success = pdb.gimp_edit_copy(preview_layer)
                    floating_sel = pdb.gimp_edit_paste(temp_preview_layer, True)
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    pdb.gimp_item_set_visible(temp_preview_layer,False) 
                else:
                    if group == "case":
                        if preview_punctuation_pos > -1:
                            preview_layer = layer_names_list[preview_punctuation_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_punctuation_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_punctuation_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_punctuation_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_punctuation_layer,False) 
                        if preview_number_pos > -1:
                            preview_layer = layer_names_list[preview_number_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_number_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_number_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_number_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_number_layer,False) 
                        if preview_lower_pos > -1:
                            preview_layer = layer_names_list[preview_lower_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_lower_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_lower_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_lower_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_lower_layer,False) 
                        if preview_upper_pos > -1:
                            preview_layer = layer_names_list[preview_upper_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_upper_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_upper_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_upper_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_upper_layer,False)
						
                    if group == "vowel":
                        if preview_consonant_pos > -1:
                            preview_layer = layer_names_list[preview_consonant_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_consonant_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_consonant_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_consonant_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_consonant_layer,False)
                        if preview_vowel_pos > -1:
                            preview_layer = layer_names_list[preview_vowel_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_vowel_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_vowel_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_vowel_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_vowel_layer,False)
							
                    if group == "shape":
                        if preview_mixed_pos > -1:
                            preview_layer = layer_names_list[preview_mixed_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_mixed_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_mixed_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_mixed_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_mixed_layer,False)
                        if preview_round_pos > -1:
                            preview_layer = layer_names_list[preview_round_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_round_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_round_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_round_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_round_layer,False)
                        if preview_straight_pos > -1:
                            preview_layer = layer_names_list[preview_straight_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_straight_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_straight_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_straight_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_straight_layer,False)
							
                pdb.gimp_image_resize_to_layers(preview_image)
            
            if do_preview:
                ####
                num_preview_layers, preview_layer_ids = pdb.gimp_image_get_layers(preview_image)
                preview_layer_names_list = preview_image.layers
                preview_loop = 0
				
                while preview_loop < num_preview_layers:
                    preview_name = pdb.gimp_item_get_name(preview_layer_names_list[preview_loop])
                    preview_name = preview_name.decode('utf-8')
					
					
                    layer_width = preview_layer_names_list[preview_loop].width
                    if layer_width > preview_image.width: layer_width = preview_image.width
                    layer_height = preview_layer_names_list[preview_loop].height
                    if layer_height > preview_image.height: layer_height = preview_image.height
					
					
                    if fixed_position: # sort out x y positions to copy from
                        random_x = 0
                        if layer_width < preview_image.width: random_x = int((preview_image.width - layer_width)/2)
                        random_y = 0
                        if layer_height < preview_image.height: random_y = int((preview_image.height - layer_height)/2)
                    else:
                        random.seed()
                        random_x = random.randrange(0, (preview_image.width + 1) - layer_width)
                        random.seed()
                        random_y = random.randrange(0, (preview_image.height + 1) - layer_height)
						
						
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(preview_image, preview_layer_names_list[preview_loop])
                    layer_copy = pdb.gimp_layer_copy(preview_layer_names_list[preview_loop], True)
                    pdb.gimp_image_insert_layer(preview_image, layer_copy, None, layer_position)
                    # set random position to copy visible layers 
                    #layer_copy.set_offsets(random_x, random_y)
					
                    # select the contents of the layer and clear contents
                    pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                    if grow>0:
                        pdb.gimp_selection_grow(preview_image, grow)
                    pdb.gimp_drawable_edit_clear(layer_copy)
					
                    # cancel use of multi decorative layers if character groups are labelled 
                    if group<>"none":
                    #if "other" in str(multi_layer_list) or "upper" in str(multi_layer_list) or "lower" in str(multi_layer_list) or "number" in str(multi_layer_list) or "punctuation" in str(multi_layer_list): 
                        many_faces = False
						
						
                    if many_faces:
                        if many_rnd: # randomize choice of faces
                            random.seed()
                            random_x = random.randrange(0, (preview_image.width + 1) - layer_width)
							
                            random_layer = random.choice(copy_multi_layer_list)
                            multi_target = multi_layer_list.index(random_layer)
							
							
                            random.seed()
					
				
                            if len(copy_multi_layer_list)>0:
                                copy_multi_layer_list.remove(random_layer)
					
                            if len(copy_multi_layer_list)<1:
                                copy_multi_layer_list = fn_copy_list(multi_layer_list)
                                random.shuffle(copy_multi_layer_list)

                        else:
                            multi_target = visi_loop % num_multi_layers
						
                        multi_loop = 0
                        while multi_loop < num_multi_layers: # set visibility of current copy layer in original image 
                            visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                            if multi_loop == multi_target:
                                pdb.gimp_item_set_visible(visible_layer, True)
                            else:
                                pdb.gimp_item_set_visible(visible_layer, False)
                            multi_loop += 1
							
                    else:
                        multi_target = 0 # topmost layer used as face for any layer not catered for by the character groups below
                        multi_loop = 0
                        while multi_loop < num_multi_layers:
                            visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                            if multi_loop == multi_target:
                                pdb.gimp_item_set_visible(visible_layer, True)
                            else:
                                pdb.gimp_item_set_visible(visible_layer, False)
                            multi_loop += 1
							
                        if "other" in str(multi_layer_list) and preview_name[0] not in punctuation_case and preview_name[0] not in number_case and preview_name[0] not in lower_case and preview_name[0] not in upper_case and preview_name[0] not in vowel_case and preview_name[0] not in consonant_case and preview_name[0] not in straight_case and preview_name[0] not in round_case and preview_name[0] not in mixed_case: 
                            counter = 0
                            while counter < num_multi_layers:
                                if "other" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
									
                                multi_loop += 1
					
                        elif preview_name[0] in upper_case and "upper" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "upper" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
								
                        elif preview_name[0] in lower_case and "lower" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "lower" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
					
                        elif preview_name[0] in number_case and "number" in str(multi_layer_list) and group == "case":
                            counter = 0
                            while counter < num_multi_layers:
                                if "number" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
                        elif preview_name[0] in punctuation_case and "punctuation" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "punctuation" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in vowel_case and "vowel" in str(multi_layer_list) and group == "vowel": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "vowel" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in consonant_case and "consonant" in str(multi_layer_list) and group == "vowel": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "consonant" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
																
														
                        elif preview_name[0] in straight_case and "straight" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "straight" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
								
                        elif preview_name[0] in round_case and "round" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "round" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in mixed_case and "mixed" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "mixed" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
					
                    # copy the contents of the visible layers into the selection
                    pdb.gimp_selection_grow(preview_image, 1)
                    success = pdb.gimp_edit_copy_visible(image)
					
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_copy, True) 
					
                    # paste the selection into the floating layer
                    floating_sel = pdb.gimp_edit_paste(layer_copy,True)
                    # and anchor it
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                    if not clear_contents: pdb.gimp_layer_set_mode(layer_copy, mode)
                    # Re-position the copied layer
                    layer_copy.set_offsets(0, 0)
					
					
                    # use the original layer to copy the alpha selection before clearing the original character
                    if clear_contents: 
					
                        pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                        if grow>0:
                            pdb.gimp_selection_grow(preview_image, grow)
                        else:
                            proc_trim_away(preview_image, layer_copy, preview_layer_names_list[preview_loop], 1)
					
					
                    # Make the original character layer visible before we merge down
					
                    pdb.gimp_item_set_visible( preview_layer_names_list[preview_loop], True)
                    # option to clear the original character
                    if clear_contents: 
                        pdb.gimp_selection_none(preview_image)
                        pdb.gimp_drawable_edit_clear(preview_layer_names_list[preview_loop])

						
						
                    if not clear_contents: pdb.gimp_layer_set_opacity(layer_copy, opacity)
                    # Merge down the copied layer to the character layer
                    pdb.gimp_image_merge_down(preview_image, layer_copy, 0)
                    # Re-set the id of the layer
                    preview_layer_names_list[preview_loop] = pdb.gimp_image_get_active_layer(preview_image)
                    # Remove annoying halo
                    #if not clear_contents: proc_trim_away(image, layer_names_list[visi_loop], layer_names_list[visi_loop], 1)
                    # Turn off layer visibility
                    pdb.gimp_item_set_visible( preview_layer_names_list[preview_loop], False) 
				
						
                    '''		
                    '''
							
					
					
                    preview_loop += 1
                    pdb.gimp_selection_none(preview_image)
				
                # Deselect
                pdb.gimp_selection_none(preview_image)
				
                # Turn on preview image first layer visibility
                pdb.gimp_item_set_visible( preview_layer_names_list[0], True) 
                pdb.gimp_image_set_active_layer(preview_image, preview_layer_names_list[0])
                # Set end undo point for preview image
                pdb.gimp_image_undo_group_end(preview_image)
				
				
            else:
                # apply to all characters in original image
                visi_loop = 0
                while visi_loop < num_layers:
                    visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                    visi_name = visi_name.decode('utf-8')
                    #if not fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): pdb.gimp_message(visi_name)
                    if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # cater for unicode symbol characters
                        # make sure that existing layers are positioned at the origin of the image
                        layer_names_list[visi_loop].set_offsets(0, 0)
                        # prepare to move layer to random position within the confines of the image size
                        layer_width = layer_names_list[visi_loop].width
                        if layer_width > im_width: layer_width = im_width
                        layer_height = layer_names_list[visi_loop].height
                        if layer_height > im_height: layer_height = im_height

                        if fixed_position: # sort out x y positions to copy from
                            random_x = 0
                            if layer_width < im_width: random_x = int((im_width - layer_width)/2)
                            random_y = 0
                            if layer_height < im_height: random_y = int((im_height - layer_height)/2)
                        else:
                            random.seed()
                            random_x = random.randrange(0, (im_width + 1) - layer_width)
                            random.seed()
                            random_y = random.randrange(0, (im_height + 1) - layer_height)
					
                        # make a duplicate of the current character layer and insert
                        layer_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                        layer_copy = pdb.gimp_layer_copy(layer_names_list[visi_loop], True)
                        pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)
                        # set random position to copy visible layers 
                        layer_copy.set_offsets(random_x, random_y)
                    
				
                        # select the contents of the layer and clear contents
                        pdb.gimp_image_select_item(image, 2, layer_copy)
                        if grow>0:
                            pdb.gimp_selection_grow(image, grow)
                        pdb.gimp_drawable_edit_clear(layer_copy)
					
                        # cancel use of multi decorative layers if character groups are labelled 
                        if group<>"none":
                        #if "other" in str(multi_layer_list) or "upper" in str(multi_layer_list) or "lower" in str(multi_layer_list) or "number" in str(multi_layer_list) or "punctuation" in str(multi_layer_list): 
                            many_faces = False
                        
						
                        if many_faces:
                            if many_rnd: # randomize choice of faces
                                random.seed()
                                random_x = random.randrange(0, (im_width + 1) - layer_width)
							
                                random_layer = random.choice(copy_multi_layer_list)
                                multi_target = multi_layer_list.index(random_layer)
							
                                #multi_target = (random.randrange(0, num_multi_layers) )
							
                                random.seed()
					
				
                                if len(copy_multi_layer_list)>0:
                                    copy_multi_layer_list.remove(random_layer)
					
                                if len(copy_multi_layer_list)<1:
                                    copy_multi_layer_list = fn_copy_list(multi_layer_list)
                                    random.shuffle(copy_multi_layer_list)

                            else:
                                multi_target = visi_loop % num_multi_layers
							
                            multi_loop = 0
                            while multi_loop < num_multi_layers: # set visibility of current copy layer
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
                        else:
                            multi_target = 0 # topmost layer used as face for any layer not catered for by the character groups below
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
                            if "other" in str(multi_layer_list) and visi_name[0] not in punctuation_case and visi_name[0] not in number_case and visi_name[0] not in lower_case and visi_name[0] not in upper_case and visi_name[0] not in vowel_case and visi_name[0] not in consonant_case and visi_name[0] not in straight_case and visi_name[0] not in round_case and visi_name[0] not in mixed_case: 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "other" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
						
                            elif visi_name[0] in upper_case and "upper" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "upper" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in lower_case and "lower" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "lower" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in number_case and "number" in str(multi_layer_list) and group == "case":
                                counter = 0
                                while counter < num_multi_layers:
                                    if "number" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in punctuation_case and "punctuation" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "punctuation" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in vowel_case and "vowel" in str(multi_layer_list) and group == "vowel": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "vowel" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in consonant_case and "consonant" in str(multi_layer_list) and group == "vowel": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "consonant" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
														
                            elif visi_name[0] in straight_case and "straight" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "straight" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in round_case and "round" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "round" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in mixed_case and "mixed" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "mixed" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
					
                        ####temp marker
                        # copy the contents of the visible layers into the selection
                        pdb.gimp_selection_grow(image, 1)
                        success = pdb.gimp_edit_copy_visible(image)
				
                        # turn on layer visibility
                        pdb.gimp_item_set_visible(layer_copy, True) 
					
                        # paste the selection into the floating layer
                        floating_sel = pdb.gimp_edit_paste(layer_copy,True)
                        # and anchor it
                        pdb.gimp_floating_sel_anchor(floating_sel)
                        # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                        if not clear_contents: pdb.gimp_layer_set_mode(layer_copy, mode)
                        # Re-position the copied layer
                        layer_copy.set_offsets(0, 0)
					
                        # use the original layer to copy the alpha selection before clearing the original character
                        if clear_contents: 
					
                            pdb.gimp_image_select_item(image, 2, layer_copy)
                            if grow>0:
                                pdb.gimp_selection_grow(image, grow)
                            else:
                                proc_trim_away(image, layer_copy, layer_names_list[visi_loop], 1)
					
					
                        # Make the original character layer visible before we merge down
					
					
					
					
                        pdb.gimp_item_set_visible( layer_names_list[visi_loop], True)
                        # option to clear the original character
                        if clear_contents: 
                            pdb.gimp_selection_none(image)
                            pdb.gimp_drawable_edit_clear(layer_names_list[visi_loop])
						
						
						
                        if not clear_contents: pdb.gimp_layer_set_opacity(layer_copy, opacity)
                        # Merge down the copied layer to the character layer
                        pdb.gimp_image_merge_down(image, layer_copy, 0)
                        # Re-set the id of the layer
                        layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
                        # Remove annoying halo
                        #if not clear_contents: proc_trim_away(image, layer_names_list[visi_loop], layer_names_list[visi_loop], 1)
                        # Turn off layer visibility
                        pdb.gimp_item_set_visible(layer_names_list[visi_loop], False) 
				
				
                
                    visi_loop += 1
				
                    if has_a_space:
                        # re-position the space layer (just in case)
                        layer_names_list[space_loop].set_offsets(0, 0)
		
                    # Deselect
                    pdb.gimp_selection_none(image)
				
                ######
				
            visi_loop = 0
            # Re-set the list of layer ids
            layer_names_list = image.layers
        
            while visi_loop < num_layers and other_layers > 0:
                visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                visi_name = visi_name.decode('utf-8')
                if visi_loop <> first_letter_layer:
                    # turn off visibility of letter layers
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
                visi_loop += 1
				
               
            # turn on visibility of character at the top of the layer stack and make it the active layer
            pdb.gimp_item_set_visible(layer_names_list[first_letter_layer],True)
            pdb.gimp_image_set_active_layer(image, layer_names_list[first_letter_layer])
				
        else:
            message_text = "There appears to be no suitable layer to create new letters from."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)            
			
		

    if xcffilename == "": do_merge_files = False
	
	
    if do_merge_files:
        cf_layers, visi_loop = 0, 0
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_names_list = image.layers
        while visi_loop < num_layers:# loop through layers and store any non-character layers for creating new faces
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # check for custom font layer now allows unicode symbols
                cf_layers += 1
                if cf_layers>0: 
                    visi_loop = num_layers                    		
            visi_loop += 1
        if cf_layers == 0: 
            do_merge_files = False
            message_text = "There appears to be no suitable character layers."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)     
	
    if do_merge_files:
        mode = fn_get_mode(old_mode, new_mode)
        font_name_pointer, file_dummy = -1, 0
        fontimg = pdb.gimp_xcf_load(file_dummy,xcffilename,xcffilename)
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop, first_letter_layer = 0, -1
		
		
	
        if do_preview:
            my_first_letter_layer = pdb.gimp_image_get_active_layer(image)
            preview_other, preview_case, preview_shape, preview_vowel = 0, 0, 0, 0
		
            preview_im_width, preview_im_height = image.width, image.height
            preview_image = pdb.gimp_image_new(preview_im_width, preview_im_height, 0)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_im_width, preview_im_height, 1, "temp preview layer", 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            pdb.gimp_image_remove_layer(preview_image, temp_preview_layer)
            preview_display = pdb.gimp_display_new(preview_image)
            # Set start undo point for preview image
            pdb.gimp_image_undo_group_start(preview_image)	
            preview_layer_pos = -1
			
			
            visi_loop = 0
            real_name_layer_list = ["dummy"] # list to hold actual character names 
            while visi_loop < num_layers: # turn off visibility of letter layers
                real_visi_name = pdb.gimp_item_get_name(layer_list[visi_loop])
                visi_name = real_visi_name.decode('utf-8')
                if visi_name == "space.png" or visi_name == " .png": 
                    has_a_space = True
                    space_loop = visi_loop
                if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                    pdb.gimp_item_set_visible(layer_list[visi_loop],False) 
                    if real_visi_name<>"space.png" and real_visi_name<>" .png": 
                        real_name_layer_list.append(real_visi_name)
                        if preview_layer_pos == -1: preview_layer_pos = visi_loop
					
                visi_loop += 1
				
				
            preview_layer = layer_list[preview_layer_pos]
            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            success = pdb.gimp_edit_copy(preview_layer)
            floating_sel = pdb.gimp_edit_paste(temp_preview_layer, True)
            pdb.gimp_floating_sel_anchor(floating_sel)
            pdb.gimp_item_set_visible(temp_preview_layer,False) 
			
			
            custom_layer = pdb.gimp_image_get_layer_by_name(fontimg, preview_layer_name)
			
            if custom_layer >= 0: #layer found
                pdb.gimp_item_set_visible(temp_preview_layer,True)
                pdb.gimp_image_set_active_layer(preview_image, temp_preview_layer)
                success = pdb.gimp_edit_copy(custom_layer)
				
				
                if fixed_position: # copies directly onto existing layer with same character name
                    floating_sel = pdb.gimp_edit_paste(temp_preview_layer, False)
                    pdb.gimp_layer_set_opacity(floating_sel, opacity)
                else: # creates a new layer to copy to 
                    temp_layer = gimp.Layer(preview_image, "Temp Layer", custom_layer.width, custom_layer.height, RGBA_IMAGE, 100, 0) # layer 
                    preview_image.add_layer(temp_layer, -1)
                    floating_sel = pdb.gimp_edit_paste(temp_layer, False)

				
                pdb.gimp_floating_sel_anchor(floating_sel)
				
                if not clear_contents: 
                    pdb.gimp_layer_set_mode(temp_layer, mode)	
				
                if not fixed_position:
                    if add_drop: 
                        #pdb.gimp_image_set_active_layer(image, drop_layer)
                        temp_layer = proc_dropshadow(preview_image, layer_list[loop], temp_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour)
                    pdb.gimp_layer_set_opacity(temp_layer, opacity)
                    real_name_layer_list[loop] = pdb.gimp_image_merge_down(preview_image, temp_layer, 2)
                pdb.gimp_item_set_visible(preview_layer,False)
                #pdb.gimp_item_set_visible(preview_layer,True)
		
            # Set end undo point for preview image
            pdb.gimp_image_undo_group_end(preview_image)
            if my_first_letter_layer >-1:pdb.gimp_item_set_visible(my_first_letter_layer,True)
        else:
		
            while loop < num_layers:
                layer_name = pdb.gimp_item_get_name(layer_list[loop])
                custom_layer = pdb.gimp_image_get_layer_by_name(fontimg, layer_name)
                if (len(layer_name) == 5  and ".png" in layer_name) or (len(layer_name) == 6 and ".png" in layer_name):
                    if first_letter_layer <0: first_letter_layer = loop
                if custom_layer >= 0: #layer found
                    pdb.gimp_item_set_visible(layer_list[loop],True)
                    pdb.gimp_image_set_active_layer(image, layer_list[loop])
                    success = pdb.gimp_edit_copy(custom_layer)
                    if fixed_position:
                        floating_sel = pdb.gimp_edit_paste(layer_list[loop], False)
                        pdb.gimp_layer_set_opacity(floating_sel, opacity)
					
                    else:
                        temp_layer = gimp.Layer(image, "Temp Layer", custom_layer.width, custom_layer.height, RGBA_IMAGE, 100, 0) # layer 
                        image.add_layer(temp_layer, -1)
                        floating_sel = pdb.gimp_edit_paste(temp_layer, False)
                    #success = pdb.gimp_edit_copy(custom_layer)
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                    if not clear_contents: 
                        pdb.gimp_layer_set_mode(temp_layer, mode)	
                    
                    pdb.gimp_image_select_item(fontimg, 2, custom_layer)
                    is_empty = pdb.gimp_selection_is_empty(fontimg)
                    pdb.gimp_selection_none(fontimg)
                    if not fixed_position: 
                        #if add_drop and "space" not in layer_name and not is_empty: 
                        if add_drop and (not is_empty): 
                            #pdb.gimp_image_set_active_layer(image, drop_layer)
                            temp_layer = proc_dropshadow(image, layer_list[loop], temp_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour)
                        pdb.gimp_layer_set_opacity(temp_layer, opacity)
                        layer_list[loop] = pdb.gimp_image_merge_down(image, temp_layer, 2)
                    pdb.gimp_item_set_visible(layer_list[loop],False)
                loop += 1
        
            if first_letter_layer >-1:pdb.gimp_item_set_visible(layer_list[first_letter_layer],True)
        
	
    temp = pdb.gimp_displays_flush()
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
		
 	
#####################################################################################################

def proc_dropshadow(im, original_layer, target_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour):
    #REM Create a text drop shadow effect - optionally clipped to the text layer for transparent type texts
    pdb.gimp_image_freeze_layers(im)
    # make a copy of the image width/height
    old_im_width, old_im_height = im.width, im.height
    name = pdb.gimp_item_get_name(target_layer)
    # Copy layer to use as the drop shadow
    shadow_layer = pdb.gimp_layer_copy(target_layer, False)
    # Add and re-position the layer in the layer stack
    im.add_layer(shadow_layer, -1)
    pdb.gimp_image_lower_item(im, shadow_layer)
    # get layer offsets for the shadow layer (which will also be the same as the original layer holding the user text)
    shadow_offset_x, shadow_offset_y = pdb.gimp_drawable_offsets(shadow_layer)
    # work out the centre point of the shadow/original layer
    shad_centre_x = (shadow_layer.width/2)
    shad_centre_y = (shadow_layer.height/2)
    # make up new size for the shadow layer to accommodate blurred result - currently twice the existing width and height
    new_shad_width = shadow_layer.width*2
    new_shad_height = shadow_layer.height*2
    # work out new offsets for the new size of the shadow layer to ensure that the shadow and original layers are still aligned
    new_shad_layer_offset_x = 0 - ((shadow_layer.width - new_shad_width)/2)
    new_shad_layer_offset_y = 0 - ((shadow_layer.height - new_shad_height)/2)
    # and apply the new offsets to the shadow layer
    pdb.gimp_layer_resize(shadow_layer, new_shad_width, new_shad_height, new_shad_layer_offset_x, new_shad_layer_offset_y)
    # make a note of the new shadow layer offsets
    shadow_resized_offset_x, shadow_resized_offset_y = pdb.gimp_drawable_offsets(shadow_layer)	
	
    # if any of the shadow layer falls outside the image boundaries
    reset_shadow = False
    if new_shad_layer_offset_x<0 or new_shad_layer_offset_y<0 or shadow_layer.width > im.width or shadow_layer.height > im.height:
        reset_shadow = True
        # set offsets of the shadow layer to image origin
        pdb.gimp_layer_set_offsets(shadow_layer,0,0)
        # set offsets for original text layer to match
        pdb.gimp_layer_set_offsets(target_layer,0 + abs(new_shad_layer_offset_x) ,0 + abs(new_shad_layer_offset_y))
        # resize the image to accomodate the new shadow layer size
        new_im_width, new_im_height = im.width, im.height
        if shadow_layer.width > im.width: new_im_width = shadow_layer.width
        if shadow_layer.height > im.height: new_im_height = shadow_layer.height
        pdb.gimp_image_resize(im, new_im_width, new_im_height, 0, 0)
		
    # get alpha selection
    pdb.gimp_image_select_item(im, 2, shadow_layer)
    # fill the text on the shadow layer with the background colour
    if drop_colour<3:
        pdb.gimp_drawable_edit_fill(shadow_layer,drop_colour-1)
    # deselect
    pdb.gimp_selection_none(im)
    #pdb.gimp_image_select_item(im, 2, original_layer)
    # blur shadow 
    pdb.plug_in_gauss(im, shadow_layer, drop_blur, drop_blur, 1)
    #pdb.gimp_selection_none(im)
    # crop the blurred layer
    pdb.plug_in_autocrop_layer(im, shadow_layer)
    # move the shadow layer to user position and adjust opacity
    temp = pdb.gimp_layer_translate(shadow_layer, drop_offset_x, drop_offset_y)
    pdb.gimp_layer_set_opacity(shadow_layer, drop_opacity)
    # merge and rename
    target_layer = pdb.gimp_image_merge_down(im,target_layer,0)
    pdb.gimp_item_set_name(target_layer, name)
    if reset_shadow: 
        cropped_offset_x, cropped_offset_y = pdb.gimp_drawable_offsets(target_layer)	
        pdb.gimp_layer_set_offsets(target_layer,shadow_offset_x, shadow_offset_y)
    pdb.gimp_image_resize(im, old_im_width, old_im_height, 0, 0)
    
    pdb.gimp_image_thaw_layers(im)
    return target_layer
    
######################################################################################################################################

def fn_copy_list(new_list): 
    copy_list = new_list[:] 
    return copy_list 
	
######################################################################################################################################

def fn_get_mode(old_mode, new_mode):
        # Use Legacy Modes if new Default Modes are set to Normal Mode
        temp_mode = 0
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
            elif new_mode == 2:
                temp_mode = 57
            elif new_mode == 3:
                temp_mode = 59
            elif new_mode == 4:
                temp_mode = 60
            elif new_mode == 5:
                temp_mode = 36
            elif new_mode == 6:
                temp_mode = 55
            elif new_mode == 7:
                temp_mode = 31
            elif new_mode == 8:
                temp_mode = 42
            elif new_mode == 9:
                temp_mode = 33
            elif new_mode == 10:
                temp_mode = 35
            elif new_mode == 11:
                temp_mode = 54
            elif new_mode == 12:
                temp_mode = 30
            elif new_mode == 13:
                temp_mode = 43
            elif new_mode == 14:
                temp_mode = 53
            elif new_mode == 15:
                temp_mode = 23
            elif new_mode == 16:
                temp_mode = 45
            elif new_mode == 17:
                temp_mode = 44
            elif new_mode == 18:
                temp_mode = 48
            elif new_mode == 19:
                temp_mode = 49
            elif new_mode == 20:
                temp_mode = 50
            elif new_mode == 21:
                temp_mode = 51
            elif new_mode == 22:
                temp_mode = 32
            elif new_mode == 23:
                temp_mode = 52
            elif new_mode == 24:
                temp_mode = 34
            elif new_mode == 25:
                temp_mode = 46
            elif new_mode == 26:
                temp_mode = 47
            elif new_mode == 27:
                temp_mode = 41
            elif new_mode == 28:
                temp_mode = 37
            elif new_mode == 29:
                temp_mode = 38
            elif new_mode == 30:
                temp_mode = 39
            elif new_mode == 31:
                temp_mode = 40
            elif new_mode == 32:
                temp_mode = 24
            elif new_mode == 33:
                temp_mode = 25
            elif new_mode == 34:
                temp_mode = 26
            elif new_mode == 35:
                temp_mode = 27
            elif new_mode == 36:
                temp_mode = 56
            mode = temp_mode
        else:
            if old_mode == 1:
                temp_mode = 1
            elif old_mode == 2:
                temp_mode = 10
            elif old_mode == 3:
                temp_mode = 4
            elif old_mode == 4:
                temp_mode = 16
            elif old_mode == 5:
                temp_mode = 7
            elif old_mode == 6:
                temp_mode = 9
            elif old_mode == 7:
                temp_mode = 3
            elif old_mode == 8:
                temp_mode = 17
            elif old_mode == 9:
                temp_mode = 5
            elif old_mode == 10:
                temp_mode = 19
            elif old_mode == 11:
                temp_mode = 18
            elif old_mode == 12:
                temp_mode = 6
            elif old_mode == 13:
                temp_mode = 8
            elif old_mode == 14:
                temp_mode = 20
            elif old_mode == 15:
                temp_mode = 21
            elif old_mode == 16:
                temp_mode = 22
            elif old_mode == 17:
                temp_mode = 11
            elif old_mode == 18:
                temp_mode = 12
            elif old_mode == 19:
                temp_mode = 13
            elif old_mode == 20:
                temp_mode = 14
            else:
                temp_mode = 0
            mode = temp_mode
			

        return mode
	
def proc_trim_away(img, l_copy, layer, amount):
    # only use when clearing the contents of the original character
    # set our feather settings
    our_feather = True
    our_feather_radius = 1
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # Comment out the shrink for use when clearing the contents of the original character
    # shrink the selection
    #pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(l_copy)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    #pdb.gimp_message("test = " + str(test) )
    return test
	
######################################################################################################################################


register(
    "python_fu_Custom_Font_Face_Maker_v0-11_Tabbed",
    "Provides options for making custom font faces",
    "Provides options for making custom font faces",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Face Maker v0.11 Tabbed...",
    "*",
    [
        (PF_TOGGLE, "fixed_bg_option", "Fixed position \nfrom visible layers:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
        (PF_SPINNER, "pf_grow", "Grow characters within layers: ", 0.0,(0.0,10.0,0.2),{'newLine':'1','tab':0}),
        (PF_TOGGLE, "multi_bg_option", "Use all available\n decorative layers:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
        (PF_TOGGLE, "multi_rnd_option", "Randomise use of\n decorative layers:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
        (PF_OPTION, "group_option", "Character groups", 0, ("None used", "By case (upper, lower, number, punctuation)", "By shape (straight, curved, mixed)", "By vowel (vowel, consonant)"), {'newLine':'1','tab':0}),
        (PF_TOGGLE, "clear_face_option", "Clear contents of \noriginal character:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
        (PF_OPTION, "mode_option", "Legacy Mode \nto use with font face:", 0, ("Normal", "Dissolve", "Lighten", "Screen", "Dodge", "Addition", "Darken", "Multiply", "Burn", "Overlay", "Softlight", "Hardlight", "Difference", "Subtract", "Grain Extract", "Grain Merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSV)"),{'newLine':'1','tab':0}),
        (PF_OPTION, "mode_option2", "New Mode \nto use with font face:", 0, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':0}),
        (PF_SPINNER, "pf_opacity", "Opacity: ", 100,(5,100,1),{'newLine':'1','tab':0}),
		
        (PF_TOGGLE, "pf_usecustomfont", "Combine the current character \nwith one from the custom font below:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
        (PF_FILENAME, "pf_names_xcffilename", ".xcf custom font file:", "", 0,{'newLine':'1','tab':1}),
        (PF_SLIDER, "pf_drop_offset_x", "Combined shadow offset x position: ", 8, (-50, 50, 1),{'newLine':'1','tab':1}),
		
        (PF_SLIDER, "pf_drop_offset_y", "Combined shadow offset y position: ", 8, (-50, 50, 1),{'newLine':'1','tab':1}),
        (PF_SLIDER, "pf_drop_blur", "Combined shadow blur percentage: ", 15, (0, 50, 1),{'newLine':'1','tab':1}),
        (PF_SLIDER, "pf_drop_opacity", "Combined shadow opacity: ", 70, (20, 100, 1),{'newLine':'1','tab':1}),
        (PF_OPTION, "drop_colour", "Combined shadow type", 0, ( "No Shadow", "Foreground Colour", "Background colour", "Layer copy"),{'newLine':'1','tab':1}),
        (PF_TOGGLE, "do_preview", "Preview characters\n(to a new image): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
		
        (PF_TAB    , "ptab"    , "tabs", 0, (_('New face options'),_('Combine & preview options')) ),
    ],
    [],
    do_stuff)

main()

	

