#!/usr/bin/env python
# -*- coding: utf-8 -*-
#



import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################

# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Sheen N Shine vx.xx...'
#
# Change Log
# Version 0.01 Provision added to call Shine Filter v 0.01 or v 0.01b by Graechan and Chisel or Carve Filter version 1.3 by Rob Antonishen
# Version 0.02 Outline width set to zero will allow the filter to add shine and/or chisel to existing custom font layers without adding outlines when outline width set to zero
#              Distort will also work on all these outlines apart from the brushed outline when outline width set to zero and Distort selection spread value > 0
# Version 0.03 Trimmed pixels from letter faces to remove slight halo effect
# Version 0.04 Added gaussian blur after chisel effect to smooth artifacts using pdb.plug_in_sel_blur with options for blur radius and blur delta
# Version 0.05 Removed dependency on shine and chisel plug-ins and replaced calls with embedded code
#              Added option to colour the shiny gloss

# Current Version of Custom Font Outlines 0.05
# Version 0.01 Effects Only
#              Separated from Outlines to leave only Shine and Chisel or Carve effects to reduce dialog height.
# Version 0.02 Effects Only
#              Added Emboss and Bumpmap Options
# Version 0.03 Effects Only
#              Added separate options for effects
# Version 0.04 Effects Only
#              Added trim option with shine effect for user to remove pixels from the character borders
# Version 0.05 Effects Only
#              Adjusted settings for bevel effect (bevel grey used and mode opacity increased to full)
# Version 0.06 Effects Only
#              Added option to offset shine; fixed bevel depth dialog setting error
# Version 0.07 Effects Only
#              Tweaked emboss - removed alpha selection during first gaussian blur and embossing before re-applying selection
# Version 0.08 Effects Only
#              Updated bumpmap
# Version 0.09 Effects Only
#              Fixed error in use of selective gaussian blur plug-in in bevel option, accepts unicode symbol characters 
# Migrated to
# Sheen N Shine
# Version 0.01 Offers two effects i.e. the existing Shine effect and new Sheen effect - a sort of reflected gloss
#			   Sheen can be used with Shine or separately takes colour from currently set bg colour
# Version 0.02 Increased range of modes and density of sheen 
# Version 0.03 Added slider to adjust emboss darkness
# Version 0.04 Maintain shine with higher emboss widths 
# Version 0.05 selection process adjusted
# Version 0.06 added emboss tweak option to blur edges of character before applying shine
#              fixed empty selection bug
# Version 0.07 further tweaking of shine
# Version 0.08 Even more revision of shine effect
# Tabbed       added option to increase depth of shine
#              added option to modify shine to include high gloss style effect
#              added options to alter opacity and blur for shine
#              changed emboss tweak to embossed edge blur
#              fixed trim option so that it is the final action thus removing any unwanted halo effect from characters


def do_stuff(image, layer,  shine_offset, gloss_colour, shine_depth, shine_modifier, shine_opacity, shine_blur, shine_mode_ptr,
            emboss_width, tweak_emboss, emboss_amount, effect_elevation, effect_azimuth, blur_radius, blur_delta,
            trim_value, use_sheen,  add_sheen, sheen_opacity, sheen_mode_ptr, do_preview, preview_char, ptab):
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    use_shine = True
	
    #use_clipboard_as_brush, do_layer_info, do_layer_names, do_path_names, do_sort = 0,0,0,0,0
    outline_colour = pdb.gimp_context_get_foreground()
	
    # Define some colours for proc chisel
    black = gimpcolor.RGB(0, 0, 0)
    white = gimpcolor.RGB(255, 255, 255)
    darkgrey = gimpcolor.RGB(50, 50, 50)
    midgrey = gimpcolor.RGB(125, 125, 125)
    grey_value = int(150 - emboss_amount)
    grey = gimpcolor.RGB(grey_value, grey_value, grey_value)
    #grey = black

    polish_colour = pdb.gimp_context_get_background()
	
    # set non UI shine plugin parameters
    shine_shadow_X, shine_shadow_Y, shine_shadow_size, shine_shadow_opacity, shine_keep_selection, shine_keep_layers = 0,0,0,0,0,0
    use_texture = False
	
    if use_sheen: use_shine = False
	
    # set non UI chisel plugin parameters
    chisel_depth, chisel_location, chisel_keep_bump = 60, 0, 0

    loop_ptr = 1
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
	
    visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
    bumpmap_layer = "none"

    if use_texture:
        while visi_loop < num_layers:
            # check for layers to create bumpmap from
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if len(visi_name) == 1 or visi_name == "space.png" or len(visi_name) == 5 and ".png" in visi_name:
                if first_letter_layer < 0: first_letter_layer = visi_loop
            else:
                pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
                is_empty = pdb.gimp_selection_is_empty(image)
                if is_empty == False: 
                    other_layers += 1
                    if other_layers == 1: 
                        # Use the top-most non-character layer as the source for the bumpmap layer and make a copy of it
                        bumpmap_layer = layer_names_list[visi_loop].copy()
                        image.add_layer(bumpmap_layer, -1)
                        bumpmap_layer_name = "Bump"
                        pdb.gimp_image_set_active_layer(image, bumpmap_layer)
                        pdb.gimp_selection_none(image)
                        # desaturate the newly created bumpmap layer
                        pdb.gimp_drawable_desaturate( bumpmap_layer, 3)
                    		
				
            visi_loop += 1
        # Update the lists of layers
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
    if other_layers == 0: use_texture = False # if there are no layers to use for the bumpmap then cancel it
	
    first_char_layer_name = "none"
	
    if preview_char == "": 
        preview_char = "A"
    else:
        preview_char = preview_char[0]
    preview_char = preview_char + ".png"
	
    if not (use_sheen==False and use_shine == False):
        # turn off visibility of layers
        visi_loop, first_char_layer, prev_check  = 0, -1, False
        while visi_loop < num_layers:
	        # get info about layer we are working on
            layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            layer_name = layer_name.decode('utf-8')
            if len(layer_name) == 1 or len(layer_name) == 5 and ".png" in layer_name:
                if preview_char in layer_name: 
                    prev_check = True
                    prev_pos = str(layer_names_list).find(preview_char)
                    prev_loop = str(layer_names_list).count(",",0,prev_pos)
                    #pdb.gimp_message("preview_char " + str(preview_char) +" is available at prev_pos "  + str(prev_pos ) + " and at prev_loop " + str(prev_loop) )
					
                if first_char_layer < 0: 
                    first_char_layer = visi_loop
                    first_char_layer_name = layer_name
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)
            visi_loop += 1
			
    if not prev_check:
            
        visi_loop = 0
        while pdb.gimp_item_get_name(layer_names_list[visi_loop]) == "space.png":
            visi_loop += 1
        preview_char = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        prev_loop = visi_loop
        #pdb.gimp_message("using alt preview_char " + str(preview_char) +" is available at prev_loop " + str(prev_loop) )
	

        preview_char = preview_char.decode('utf-8')	######## Added to catch higher value unicodes
	
        #pdb.gimp_message("preview_char = " + str(preview_char) + " do_preview = " + str(do_preview) )
		
    if do_preview:
		
        # copy the preview character and paste it as a  new image
        preview_layer = pdb.gimp_image_get_layer_by_name(image, preview_char)
        success = pdb.gimp_edit_copy(preview_layer)
        preview_image = pdb.gimp_edit_paste_as_new_image(preview_layer)
            
        pdb.gimp_image_undo_group_start(preview_image)
        preview_display = pdb.gimp_display_new(preview_image)

        preview_layer_names_list = preview_image.layers
		
		
        preview_layer_id = preview_layer_names_list[0]
        preview_layer_name = pdb.gimp_item_get_name(preview_layer_names_list[0])
        preview_layer_name = preview_char.decode('utf-8')	######## Added to catch higher value unicodes
		
		
        test_for_custom_layer = True
			
        display_layer_name =  pdb.gimp_item_get_name(preview_layer_names_list[0])
        display_layer_name = display_layer_name.decode('utf-8')	
		
        # check we have the layer id of the currently active layer
        pdb.gimp_image_set_active_layer(preview_image, preview_layer_id)
        layer_id = pdb.gimp_image_get_active_layer(preview_image)
        active_layer_id = pdb.gimp_image_get_active_layer(preview_image)
		
		
        if test_for_custom_layer:

            # ensure we have the active layer and make it visible
            #active_layer_id = pdb.gimp_image_get_active_layer(image)
            pdb.gimp_item_set_visible(active_layer_id, True)

            if use_sheen:
                active_layer = preview_image.active_layer 
                active_layer_id = proc_sheen(preview_image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, sheen_mode_ptr, add_sheen, emboss_width, sheen_opacity)	
				
		
            if use_shine:
                if tweak_emboss>0: 
                    #pdb.gimp_message("in tweak_emboss")
                    g_blur = active_layer_id.height/200
                    if g_blur<1: g_blur = 1
                    # select the contents of the layer
                    pdb.gimp_image_select_item(preview_image, 2, active_layer_id)
                    # shrink the selection
                    pdb.gimp_selection_shrink(preview_image, tweak_emboss)
                    # invert selection
                    pdb.gimp_selection_invert(preview_image)
                    pdb.plug_in_gauss(preview_image, active_layer_id, g_blur, g_blur, 0)
                    # deselect
                    pdb.gimp_selection_none(preview_image)
                active_layer = preview_image.active_layer 
                active_layer_id = proc_gloss(preview_image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, shine_mode_ptr, add_sheen, emboss_width, sheen_opacity, grey, tweak_emboss, shine_depth, shine_modifier, shine_opacity, sheen_mode_ptr, shine_blur)
  
						
            if trim_value>0: proc_trim_away(preview_image, active_layer_id, trim_value)# shrink selection boundary by set amount of pixels
		
        pdb.gimp_image_undo_group_end(preview_image)
    else:
	
			
        while loop >= 0:
	        # get info about layer we are working on
            layer_id = layer_ids[loop]
            layer_name = layer_names_list[loop]
	
            # cater for unicode characters above normal range
            display_layer_name =  pdb.gimp_item_get_name(layer_names_list[loop])
            display_layer_name = display_layer_name.decode('utf-8')	
	
            # check we have the layer id of the currently active layer
            pdb.gimp_image_set_active_layer(image, layer_name)
            layer_id = pdb.gimp_image_get_active_layer(image)

            # check to see if the current layer is a custom font character layer
            test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
		
            # if the outline width has been set to zero but no chisel or shine has been set don't do it
            #if use_chisel == False and use_shine == False: test_for_custom_layer = False
		
            if test_for_custom_layer:

                # ensure we have the active layer and make it visible
                active_layer_id = pdb.gimp_image_get_active_layer(image)
                pdb.gimp_item_set_visible(active_layer_id, True)

                if use_sheen:
                    active_layer = image.active_layer 
                    active_layer_id = proc_sheen(image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, sheen_mode_ptr, add_sheen, emboss_width, sheen_opacity)	
				
                if use_shine:
                    if tweak_emboss>0: 
                        g_blur = active_layer_id.height/200
                        if g_blur<1: g_blur = 1
                        # select the contents of the layer
                        pdb.gimp_image_select_item(image, 2, active_layer_id)
                        # shrink the selection
                        pdb.gimp_selection_shrink(image, tweak_emboss)
                        # invert selection
                        pdb.gimp_selection_invert(image)
                        pdb.plug_in_gauss(image, active_layer_id, g_blur, g_blur, 0)
                        # deselect
                        pdb.gimp_selection_none(image)
                    active_layer = image.active_layer 
                    active_layer_id = proc_gloss(image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, shine_mode_ptr, add_sheen, emboss_width, sheen_opacity, grey, tweak_emboss, shine_depth, shine_modifier, shine_opacity, sheen_mode_ptr, shine_blur)
  
						
                    if trim_value>0: proc_trim_away(image, active_layer_id, trim_value)# shrink selection boundary by set amount of pixels
				              
				
                if (num_layers - loop_ptr) > 0: pdb.gimp_item_set_visible(active_layer_id, False)
			
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
			
            loop_ptr += 1
            loop = num_layers - loop_ptr
        

    if first_char_layer_name <> "none":
        first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
        pdb.gimp_item_set_visible(first_char_layer_id, True)
        pdb.gimp_image_set_active_layer(image,first_char_layer_id)
		
		
    pdb.gimp_displays_flush()
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
		
#### End of do_stuff

######################################################################################################################################

def proc_do_bump(image, other_layers, bumpmap_layer_id, bumpmap_layer_name, active_layer, black, white, bumpmap_width, waterlevel, ambient, effect_elevation, effect_azimuth, effect_mode_ptr):
    if other_layers > 0:
        work_layer = active_layer
        work_layer_name = work_layer.name
        width = work_layer.width
        height = work_layer.height
        bumpmap_width = bumpmap_width * 0.65
        if bumpmap_width < 1: bumpmap_width = 1
		
        if waterlevel > 0:
            waterlevel = waterlevel/100
            waterlevel = 1.0 - waterlevel
        else:
            waterlevel = 0.0
			
        if ambient > 0:
            ambient = ambient/100
            ambient = 1.0 - ambient
        else:
            ambient = 0.0        

        pdb.plug_in_bump_map(image, work_layer, bumpmap_layer_id, effect_azimuth, effect_elevation, bumpmap_width, 0, 0, waterlevel, ambient, True, False, 0)
		
        active_layer = image.active_layer 
        return active_layer
        
######################################################################################################################################
	
def proc_emboss(image, active_layer, black, white, effect_width, effect_elevation, effect_azimuth, emboss_blur, effect_mode_ptr, blur_radius, blur_delta):
    if effect_width < 1: effect_width = 1
    grey = gimpcolor.RGB(100, 100, 100)
    work_layer = active_layer # ie character layer
    work_layer_name = work_layer.name # ie character layer id
    width = work_layer.width
    height = work_layer.height
    colour = gimp.get_background()
    emboss_mode = fn_get_mode(effect_mode_ptr)
	
    layer_blur = image.height * 0.015
    if layer_blur < 1: layer_blur = 1	
	
    pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_background(white)
	
    pdb.gimp_image_select_item(image, 2, work_layer)
	
    emboss_layer = gimp.Layer(image, "Emboss Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the bump map
    image.add_layer(emboss_layer, -1)
    emboss_layer.fill(1)
    pdb.gimp_drawable_edit_fill(emboss_layer, 0)
	
    #grey_layer = gimp.Layer(image, "Grey Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the bump map
    #image.add_layer(grey_layer, -1)
    #pdb.gimp_context_set_foreground(grey)	
    #grey_layer.fill(1)
    #pdb.gimp_drawable_edit_fill(grey_layer, 0)
	
    pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_background(white)
	
    pdb.gimp_selection_none(image) ####
    pdb.plug_in_gauss_iir(image, emboss_layer, emboss_blur, True, True) # blur the bump map
	
    other_layers = 1
    bumpmap_layer_name = "Emboss Layer"
    bumpmap_layer = emboss_layer
	
    #active_layer = new_layer
	
    #temp values
    bump_waterlevel = 0
    bump_ambient = 0
    waterlevel = 0
    ambient = 0
    #end temp values
	
    bump_depth = effect_width
	
    work_layer = active_layer
    work_layer_name = work_layer.name
    width = work_layer.width
    height = work_layer.height
    #bumpmap_width = bumpmap_width * 0.65
    #if bumpmap_width < 1: bumpmap_width = 1
		
    if waterlevel > 0:
        waterlevel = waterlevel/100
        waterlevel = 1.0 - waterlevel
    else:
        waterlevel = 0.0
			
    if ambient > 0:
        ambient = ambient/100
        ambient = 1.0 - ambient
    else:
        ambient = 0.0        

    pdb.plug_in_bump_map(image, work_layer, emboss_layer, effect_azimuth, effect_elevation, bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
    #work_layer = image.active_layer
	
    image.remove_layer(emboss_layer)
    #image.lower_layer(grey_layer)
    #temp values
    #blur_radius = 10
    #blur_delta = 8
    #end temp values
    pdb.gimp_image_select_item(image, 2, work_layer)
    ##pdb.plug_in_sel_gauss(image, work_layer, blur_radius, blur_delta)
    #pdb.plug_in_gauss_iir(image, grey_layer, 5, True, True) # blur the bump map
    #pdb.plug_in_gauss_iir2(image, grey_layer, blur_radius, blur_radius) # blur the bump map
    #pdb.plug_in_gauss_rle(image, grey_layer, blur_radius, True, True) # blur the bump map
    #work_layer.mode = emboss_mode
    #pdb.gimp_image_merge_down(image, work_layer, 0)
    work_layer = image.active_layer
    #pdb.gimp_item_set_name(work_layer, work_layer_name)
    #image.remove_layer(grey_layer)
		
    pdb.gimp_selection_none(image)
    active_layer = image.active_layer 
    ##pdb.plug_in_sel_gauss(image, active_layer, 10, 255)
    pdb.plug_in_sel_gauss(image, active_layer, blur_radius, blur_delta)
    return active_layer	
	
		
######################################################################################################################################

def proc_sheen(image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, sheen_mode_ptr, add_sheen, emboss_width, sheen_opacity):
    work_layer = active_layer
    work_layer_name = work_layer.name
    width = work_layer.width
    height = work_layer.height
    elevation = 85
    #emboss_width = 10
    colour = gimp.get_background()
    polish_mode = fn_get_mode(sheen_mode_ptr)
    rotate = effect_azimuth
    effect_azimuth = 360-(effect_azimuth+90)
    if effect_azimuth>360: effect_azimuth = 360
	
    layer_blur = image.height * 0.015
    if layer_blur < 1: layer_blur = 1	
	
    sheen_size = int(max(width,height)*1.5)
    sheen_layer = gimp.Layer(image, "Sheen Layer", sheen_size, sheen_size, RGBA_IMAGE, 100, 0)
    image.add_layer(sheen_layer, -1)
    pdb.gimp_layer_set_offsets(sheen_layer, 0-int((sheen_size/2)-(width/2)), 0-int((sheen_size/2)-(height/2)))
    pdb.gimp_context_set_foreground(polish_colour)
    pdb.gimp_context_set_gradient_fg_transparent()
    pdb.gimp_drawable_edit_gradient_fill(sheen_layer, 0, 0, False, 1, 0, True, 0, 0, 0, int(sheen_size/2))
    pdb.gimp_item_transform_rotate(sheen_layer,math.radians(rotate),True, 0, 0)
    pdb.gimp_image_select_item(image, 2, work_layer)
    pdb.gimp_selection_shrink(image, 1)
    pdb.gimp_selection_invert(image)
    pdb.gimp_edit_clear(sheen_layer)
    pdb.gimp_selection_none(image)
    sheen_layer.mode = polish_mode
    pdb.gimp_layer_set_opacity(sheen_layer, sheen_opacity)
    work_layer = pdb.gimp_image_merge_down(image, sheen_layer, 2)
	
    new_layer = image.active_layer
    active_layer = image.active_layer 
    return active_layer

    #### End of proc_sheen
	
######################################################################################################################################

def proc_gloss(image, active_layer, black, white, effect_azimuth, gloss_colour, shine_offset, polish_colour, shine_mode_ptr, add_sheen, emboss_width, sheen_opacity, grey, tweak_emboss, shine_depth, shine_modifier, shine_opacity, sheen_mode_ptr, shine_blur):
    # Shine effect 
    work_layer = active_layer
    work_layer_name = work_layer.name
    width = work_layer.width
    height = work_layer.height
    elevation = 85
    emboss_width = 10
    colour = gimp.get_background()
    shine_mode = fn_get_mode(shine_mode_ptr)
    polish_mode = fn_get_mode(sheen_mode_ptr)
	
    rotate = effect_azimuth
    layer_blur = image.height * 0.015
    if layer_blur < 1: layer_blur = 1	
	
    if add_sheen:
        sheen_size = int(max(width,height)*1.5)
        sheen_layer = gimp.Layer(image, "Sheen Layer", sheen_size, sheen_size, RGBA_IMAGE, 100, 0)
        image.add_layer(sheen_layer, -1)
        pdb.gimp_layer_set_offsets(sheen_layer, 0-int((sheen_size/2)-(width/2)), 0-int((sheen_size/2)-(height/2)))
        pdb.gimp_context_set_foreground(polish_colour)
        pdb.gimp_context_set_gradient_fg_transparent()
        pdb.gimp_drawable_edit_gradient_fill(sheen_layer, 0, 0, False, 1, 0, True, 0, 0, 0, int(sheen_size/2))
        pdb.gimp_item_transform_rotate(sheen_layer,math.radians(rotate),True, 0, 0)
        pdb.gimp_image_select_item(image, 2, work_layer)
        pdb.gimp_selection_shrink(image, 1)
        pdb.gimp_selection_invert(image)
        pdb.gimp_edit_clear(sheen_layer)
        pdb.gimp_selection_none(image)
        sheen_layer.mode = polish_mode
        pdb.gimp_layer_set_opacity(sheen_layer, sheen_opacity)
        work_layer = pdb.gimp_image_merge_down(image, sheen_layer, 2)
        #new_layer = pdb.gimp_image_merge_down(image, sheen_layer, 2)
	
    #pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_foreground(grey)
    pdb.gimp_context_set_background(white)
	
	
    pdb.gimp_image_select_item(image, 2, work_layer)
	
    #new_layer = gimp.Layer(image, "Shine Layer", width, height, RGBA_IMAGE, 100, 0)
    new_layer = gimp.Layer(image, "Shine Layer", width, height, RGBA_IMAGE, 100, 0)
    image.add_layer(new_layer, -1)
    new_layer.fill(1) # fill layer with white background colour
    pdb.gimp_drawable_edit_fill(new_layer, 0) # fill selected region with black foreground colour

    pdb.gimp_image_set_active_layer(image, new_layer)
    pdb.gimp_item_set_name(work_layer,"Original layer")
	
	
	
    # apply shine
    #pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_foreground(grey)
    pdb.gimp_context_set_background(white)
	
    pdb.plug_in_gauss_rle2(image, new_layer, layer_blur, layer_blur) # prepare layer with blur in order for emboss to work
    pdb.plug_in_emboss(image, new_layer, effect_azimuth, elevation, emboss_width, 1)
    #pdb.gimp_selection_shrink(image, 1) # new
    pdb.gimp_selection_invert(image) # invert selection ready to clear surrounding pixels
    #pdb.gimp_image_select_item(image, 2, new_layer) # new
    pdb.gimp_drawable_edit_clear(new_layer)
    pdb.gimp_selection_invert(image)
	
    if shine_modifier == 0: # apply various shines
        pdb.plug_in_colortoalpha(image ,new_layer ,(254, 254, 254)) #  shine only
    elif shine_modifier == 1:
        pdb.plug_in_colortoalpha(image ,new_layer ,(253, 253, 253)) # shine with gloss
    elif shine_modifier > 1:
        pdb.plug_in_colortoalpha(image ,new_layer ,(252, 252, 252)) # shine with more opaque gloss
	
    pdb.gimp_context_set_sample_threshold_int(1) # set selection threshold
    #alpha_colour = gimpcolor.RGB(my_colour)
    #pdb.gimp_image_select_color(image, 2, new_layer, my_colour)
    if not pdb.gimp_selection_is_empty(image): # do not cut if there is not a selection
        success = pdb.gimp_edit_cut(new_layer)
    pasted_layer = gimp.Layer(image, "Pasted Layer", width, height, RGBA_IMAGE, 100, 0) # paste any selected highlight to layer
    image.add_layer(pasted_layer, -1)
    floating_sel = pdb.gimp_edit_paste(pasted_layer,True)
    pdb.gimp_floating_sel_anchor(floating_sel)
    pdb.gimp_selection_none(image)
    #pdb.gimp_edit_paste(pasted_layer, True)
    pdb.gimp_drawable_offset(pasted_layer, False, 1, shine_offset, shine_offset) # offset selected area by user amount
    
    if gloss_colour: # colour the highlight
        pdb.gimp_image_select_item(image, 2, pasted_layer)
        sel_is_empty = pdb.gimp_selection_is_empty(image) # better check that the selection still exists before shrinking
        if not sel_is_empty:
            pdb.gimp_image_select_item(image, 2, pasted_layer)
            pdb.gimp_context_set_background(colour) # get bg colour
            colour_layer = gimp.Layer(image, "Colour Layer", width, height, RGBA_IMAGE, 100, 35) # add a darken only mode layer
            pdb.gimp_selection_shrink(image, 1) # shrink selection boundary by 1 pixel
            image.add_layer(colour_layer, -1)
            colour_layer.fill(1) # fill the layer selection
            pdb.gimp_selection_none(image)
            pdb.gimp_image_merge_down(image, colour_layer, 0)
            pasted_layer = image.active_layer # reset layer id
	
    pdb.plug_in_blur(image, pasted_layer) # blur the highlight
    if shine_modifier == 0: pdb.plug_in_gauss_iir2(image, pasted_layer, shine_blur, shine_blur)
    if shine_modifier ==0: # broaden the highlight 
        pdb.plug_in_autocrop_layer(image, pasted_layer)
        if shine_depth == 0:
            temp_layer = pdb.gimp_layer_copy(pasted_layer, True)
            image.add_layer(temp_layer, -1)
            pasted_layer = pdb.gimp_image_merge_down(image, temp_layer, 0)
        if shine_depth == 1:
            temp_layer = pdb.gimp_layer_copy(pasted_layer, True)
            image.add_layer(temp_layer, -1)
            pdb.gimp_layer_translate(temp_layer,1,1)
            pasted_layer = pdb.gimp_image_merge_down(image, temp_layer, 0)
        if shine_depth == 2:
            temp_layer = pdb.gimp_layer_copy(pasted_layer, True)
            image.add_layer(temp_layer, -1)
            pdb.gimp_layer_translate(temp_layer,1,1)
            temp_layer2 = pdb.gimp_layer_copy(pasted_layer, True)
            image.add_layer(temp_layer2, -1)
            pdb.gimp_layer_translate(temp_layer2,-1,-1)
            pasted_layer = pdb.gimp_image_merge_down(image, temp_layer, 0)
            pasted_layer = pdb.gimp_image_merge_down(image, temp_layer2, 0)
    
    #pdb.gimp_drawable_offset(new_layer, False, 1, shine_offset, shine_offset)
    pdb.gimp_image_merge_down(image, new_layer, 0)
	
    pasted_layer.mode = shine_mode # apply required mode
    pdb.gimp_layer_set_opacity(pasted_layer, shine_opacity) # apply opacity
    pdb.gimp_image_merge_down(image, pasted_layer, 0)
    new_layer = image.active_layer 
    pdb.gimp_item_set_name(new_layer, work_layer_name)
		
    pdb.gimp_selection_none(image)
    new_layer = image.active_layer
    active_layer = image.active_layer 

    return active_layer

    #### End of proc_gloss
	
######################################################################################################################################

def proc_text_bevel(image, text_layer, bevel_width, bevel_depth, bevel_waterlevel, bevel_ambient, feather, black, white, midgrey, bevel_mode, bevel_azimuth, bevel_elevation, blur_radius, blur_delta):
    #### Create a bevel around text and return the new layer as bevel_layer
    #Rem Adjust values taken from dialog
    bevel_width = int(bevel_width * 2.55)
    if bevel_width < 2: bevel_width = 2
    blur_radius = blur_radius * 1
    if blur_delta>255: blur_delta = 255
	
    if bevel_ambient > 0.0:
        ambient = bevel_ambient/100
    else:
        ambient = 0.0
    if bevel_waterlevel > 0.0:
        waterlevel = bevel_waterlevel/100
    else:
        waterlevel = 0.0
		
    #Rem Select text outline and save it
    pdb.gimp_image_select_item(image, 2, text_layer)
    saved_selection = pdb.gimp_selection_save(image)

    #### Rem Make a bumpmap layer ######
    bumpmap_layer = text_layer.copy()
    image.add_layer(bumpmap_layer, -1)
    bumpmap_layer.name = text_layer.name + " Bumpmap"
    pdb.gimp_image_raise_item_to_top(image, bumpmap_layer)

    pdb.gimp_selection_feather(image, feather)
    blurred_selection = pdb.gimp_selection_save(image)
    loop = bevel_width
    pdb.gimp_selection_shrink(image, loop)

    while (pdb.gimp_selection_is_empty(image) == True ):
        loop = loop - 1
        pdb.gimp_image_select_item(image, 2, blurred_selection)
        pdb.gimp_selection_shrink(image, loop)
        
    bevel_width = min(bevel_width, loop)
    pdb.gimp_image_select_item(image, 2, blurred_selection)

    gimp.set_foreground(0, 0, 0)
    bumpmap_layer.fill(0)
	

    loop = 1
    #Rem Temp dirty fill_value 
    fill_value = (255 / bevel_width)
    while (loop <= bevel_width):

        #Rem Using Temp fill_value until something better comes to mind
        pdb.gimp_image_select_item(image, 2, blurred_selection)
        pdb.gimp_selection_shrink(image, loop -1)
        greycolour = int(fill_value)
        gimp.set_foreground(greycolour, greycolour, greycolour)

        pdb.gimp_edit_fill(bumpmap_layer, 0)

        #Rem Temp fill_value
        fill_value = fill_value + (255 / bevel_width)
        loop = loop + 1

    gimp.set_foreground(255, 255, 255)
    pdb.gimp_image_select_item(image, 2, blurred_selection)
    pdb.gimp_selection_shrink(image, bevel_width)

    if (pdb.gimp_selection_is_empty(image) == FALSE):
        pdb.gimp_edit_fill(bumpmap_layer, 0)

    pdb.gimp_selection_none(image)

    #### Make Bevel from bumpmap ####
    bevel_layer = text_layer.copy()
    image.add_layer(bevel_layer, -1)
    bevel_layer.name = text_layer.name + " Bevel"
    pdb.gimp_image_raise_item_to_top(image, bevel_layer)

    gimp.set_foreground(128, 128, 128)
    bevel_layer.fill(0)
    bevel_azimuth = bevel_azimuth - 180
    if bevel_azimuth < 0: bevel_azimuth = 360 + (bevel_azimuth)
    pdb.plug_in_bump_map(image, bevel_layer, bumpmap_layer, bevel_azimuth, bevel_elevation, bevel_depth, 0, 0, waterlevel, ambient, True, False, 0)
    bevel_layer.mode = bevel_mode
    pdb.gimp_layer_set_opacity(bevel_layer, 100)
	
    #### Delete area outside the bevel ####
    pdb.gimp_image_select_item(image, 2, saved_selection)
    pdb.gimp_selection_invert(image)

    if (pdb.gimp_selection_is_empty(image) == False):
        pdb.gimp_edit_clear(bevel_layer)

    pdb.gimp_image_select_item(image, 2, saved_selection)
    pdb.gimp_selection_shrink(image, bevel_width)

    if (pdb.gimp_selection_is_empty(image) == False):
        pdb.gimp_edit_clear(bevel_layer)
	
    pdb.gimp_image_select_item(image, 2, saved_selection)
 
    pdb.gimp_selection_none(image)
    pdb.gimp_drawable_set_visible(bumpmap_layer, False)	
    image.remove_channel(saved_selection)
    image.remove_channel(blurred_selection)
    image.remove_layer(bumpmap_layer)
    pdb.gimp_image_merge_down(image, bevel_layer, 0)
    active_layer = image.active_layer 
    pdb.plug_in_sel_gauss(image, active_layer, blur_radius, blur_delta)
    active_layer = image.active_layer 
    return active_layer
	
#### End of proc_text_bevel

######################################################################################################################################
	
def proc_trim_away(img, layer, amount):    
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################

def fn_get_mode(new_mode):
    temp_mode = 0
    n_mode_info = "Normal"
    if new_mode <> 0:
        if new_mode == 1:
            temp_mode = 1
            n_mode_info = "Dissolve"
        elif new_mode == 2:
            temp_mode = 57
            n_mode_info = "Color Erase"
        elif new_mode == 3:
            temp_mode = 59
            n_mode_info = "Merge"
        elif new_mode == 4:
            temp_mode = 60
            n_mode_info = "Split"
        elif new_mode == 5:
            temp_mode = 36
            n_mode_info = "Lighten Only"
        elif new_mode == 6:
            temp_mode = 55
            n_mode_info = "Luma Luminance Lighten"
        elif new_mode == 7:
            temp_mode = 31
            n_mode_info = "Screen"
        elif new_mode == 8:
            temp_mode = 42
            n_mode_info = "Dodge"
        elif new_mode == 9:
            temp_mode = 33
            n_mode_info = "Addition"
        elif new_mode == 10:
            temp_mode = 35
            n_mode_info = "Darken only"
        elif new_mode == 11:
            temp_mode = 54
            n_mode_info = "Luma Luminance darken"
        elif new_mode == 12:
            temp_mode = 30
            n_mode_info = "Multiply"
        elif new_mode == 13:
            temp_mode = 43
            n_mode_info = "Burn"
        elif new_mode == 14:
            temp_mode = 53
            n_mode_info = "Linear burn"
        elif new_mode == 15:
            temp_mode = 23
            n_mode_info = "Overlay"
        elif new_mode == 16:
            temp_mode = 45
            n_mode_info = "Soft light"
        elif new_mode == 17:
            temp_mode = 44
            n_mode_info = "Hard light"
        elif new_mode == 18:
            temp_mode = 48
            n_mode_info = "Vivid light"
        elif new_mode == 19:
            temp_mode = 49
            n_mode_info = "Pin light"
        elif new_mode == 20:
            temp_mode = 50
            n_mode_info = "Linear light"
        elif new_mode == 21:
            temp_mode = 51
            n_mode_info = "Hard mix"
        elif new_mode == 22:
            temp_mode = 32
            n_mode_info = "Difference"
        elif new_mode == 23:
            temp_mode = 52
            n_mode_info = "Exclusion"
        elif new_mode == 24:
            temp_mode = 34
            n_mode_info = "Subtract"
        elif new_mode == 25:
            temp_mode = 46
            n_mode_info = "Grain extract"
        elif new_mode == 26:
            temp_mode = 47
            n_mode_info = "Grain merge"
        elif new_mode == 27:
            temp_mode = 41
            n_mode_info = "Divide"
        elif new_mode == 28:
            temp_mode = 37
            n_mode_info = "Hue HSV"
        elif new_mode == 29:
            temp_mode = 38
            n_mode_info = "Saturation HSV"
        elif new_mode == 30:
            temp_mode = 39
            n_mode_info = "Color HSL"
        elif new_mode == 31:
            temp_mode = 40
            n_mode_info = "Value HSL"
        elif new_mode == 32:
            temp_mode = 24
            n_mode_info = "Hue LCH"
        elif new_mode == 33:
            temp_mode = 25
            n_mode_info = "Chroma LCH"
        elif new_mode == 34:
            temp_mode = 26
            n_mode_info = "Color LCH"
        elif new_mode == 35:
            temp_mode = 27
            n_mode_info = "Lightness LCH"
        elif new_mode == 36:
            temp_mode = 56
        n_mode_info = "Luminance"
    mode = temp_mode

    return mode
	
######################################################################################################################################

def fn_set_bevel_mode(bevel_mode_ptr):
    # Use New Modes when creating a bevel
    bevel_mode = 23
    if bevel_mode_ptr == 0:
        bevel_mode = 23 #Overlay Mode
    elif bevel_mode_ptr == 1:
        bevel_mode = 30 #Multiply Mode
    elif bevel_mode_ptr == 2:
        bevel_mode = 35 #Darken Only Mode
    elif bevel_mode_ptr == 3:
        bevel_mode = 40 #HSV Value Mode
    elif bevel_mode_ptr == 4:
        bevel_mode = 41 #Divide Mode
    elif bevel_mode_ptr == 5:
        bevel_mode = 42 #Dodge Mode
    elif bevel_mode_ptr == 6:
        bevel_mode = 44 # Hardlight Mode
    if bevel_mode_ptr == 7:
        bevel_mode = 45 # Softlight Mode
    elif bevel_mode_ptr == 8:
        bevel_mode = 46 #Grain Extract Mode
    elif bevel_mode_ptr == 9:
        bevel_mode = 47 #Grain Merge Mode
    elif bevel_mode_ptr == 10:
        bevel_mode = 48 #Vivid Mode
    elif bevel_mode_ptr == 11:
        bevel_mode = 50 #Linear Light Mode
    elif bevel_mode_ptr == 12:
        bevel_mode = 51 #Hard Mix Mode
    elif bevel_mode_ptr == 13:
        bevel_mode = 56 #Luminance Mode
    return bevel_mode


######################################################################################################################################

register(
    "python_fu_Custom_Font_Sheen_N_Shine_v0-08_Tabbed",
    "Adds shine and sheen effects to custom fonts",
    "Provides options for adding shine and sheen effects to custom fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Sheen N Shine v0.08 Tabbed...",
    "*",
    [
    (PF_SPINNER, "pf_shine_offset", "Shine offset amount: ", 0,(0,20,1),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "pf_shine_colour", "Use (BG) colour for Shine: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_shine_depth", "Shine edge depth: ", 0,(0,2,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_shine_col_mod", "Shine effect modifier: ", 0,(0,2,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_effect_opacity", "Shine opacity: ", 100,(20,100,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_highlight_blur", "Shine blur: ", 3,(0,50,1),{'newLine':'1','tab':0}),
    (PF_OPTION, "shine_mode_option2", "Shine Mode to use: ", 26, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':0}),
	
    (PF_SPINNER, "pf_emboss_width", "Emboss width ", 10,(1,50,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_tweak_emboss", "Emboss edge blur extent: ",  1,(0,5,1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_emboss_colour", "Emboss emphasis (Light - Dark): ", 50, (0, 150, 1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_effect_elevation", "Shine elevation: ", 45.0,(0.5,90,0.1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_shine_azimuth", "Shine azimuth: ", 315,(0,360,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_blur_radius", "Artifact blur radius: ", 10,(0.0,100.0,0.1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_blur_delta", "Artifact blur delta amount: ", 15,(0,255,1),{'newLine':'1','tab':1}),
	
    (PF_SPINNER, "pf_shine_trim", "Trim character in pixels: ", 1,(0,10,1), {'newLine':'1','tab':2}),
    (PF_TOGGLE, "pf_add_sheen_effect", "Add Sheen effect only: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "pf_add_sheen", "Add Sheen (BG colour) to Shine: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_SLIDER, "pf_opacity", "Sheen opacity: ", 60, (20, 100, 5), {'newLine':'1','tab':2}),
    (PF_OPTION, "sheen_mode_option2", "Sheen Mode to use: ", 26, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':2}),
    (PF_TOGGLE, "Preview_option", "Preview single character\n(to a new image): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_STRING, "preview_string", "Preview character: ", ("A"), 0,{'newLine':'1','tab':2}),

	
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Shine Options'),_('Emboss Options'),_('Sheen & Preview Options'))),
    ],
    [],
    do_stuff)

main()

	

